use crate::*;
use enumset::EnumSetType;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::fmt;
use std::io::{stdout, Write};
use std::path::PathBuf;
use std::process::Command;

#[derive(Debug, Serialize, Deserialize, EnumSetType)]
//#[enumset(serialize_as_list)] // TODO: use it or not?
enum UpdateSteps {
    Fetch,
    Compile,
    Install,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Updater {
    systems: Vec<System>,
    steps: UpdateSteps,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct System {
    //name: String,
    fetch: Option<Cmd>,
    compile: Option<Cmd>,
    install: Cmd,
    // deps or rDeps : Tree
    // exclusive_with : List
}

#[derive(Debug, Serialize, Deserialize)]
struct Cmd {
    exe: String,
    params: Vec<String>,
    current_dir: Option<PathBuf>,
    env: HashMap<String, String>,
}

impl Updater {
    fn new() -> Updater {
        let mut up = Updater {
            systems: vec![],
            steps: UpdateSteps::Fetch,
        };
        up.systems.push(System {
            fetch: None,
            compile: None,
            install: Cmd::new(),
        });
        up
    }

    /// To create a sample config from code
    #[doc(hidden)]
    fn write_config(&self, opt: &Opt) {
        use std::fs::OpenOptions;

        let mut f = OpenOptions::new()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&opt.config_file)
            .unwrap();

        f.write_all(serde_yaml::to_string(&self).unwrap().as_bytes())
            .unwrap();
    }

    pub fn parse_config(opt: &Opt) -> Result<Updater> {
        // let u = Updater::new();
        // u.write_config(opt);

        let file = std::fs::read_to_string(&opt.config_file).unwrap();
        Ok(serde_yaml::from_str(&file).unwrap())
        // TODO:
    }

    pub fn update_all(&self) -> Result<()> {
        let mut errors = vec![];
        for sys in &self.systems {
            if let Err(err) = self.update(sys) {
                eprintln!("Error catched {}", err);
                errors.push(err);
            }
        }

        // TODO:
        if errors.len() == 0 {
            Ok(())
        } else {
            Err(Error::new(
                errors,
                ErrorKind::Fetch, // TODO: Why should I choose here?
            ))
        }
    }

    fn update(&self, sys: &System) -> Result<()> {
        if self.steps == UpdateSteps::Fetch {
            sys.fetch()?;
        }
        if self.steps == UpdateSteps::Compile {
            sys.compile()?;
        }
        if self.steps == UpdateSteps::Install {
            sys.install()?;
        }

        Ok(())
    }
}

impl System {
    pub fn fetch(&self) -> Result<()> {
        if let Some(fetch) = &self.fetch {
            fetch.execute()?;
        }
        Ok(())
    }

    pub fn compile(&self) -> Result<()> {
        if let Some(compile) = &self.compile {
            compile.execute()?;
        }
        Ok(())
    }

    pub fn install(&self) -> Result<()> {
        self.install.execute()?;
        Ok(())
    }
}

impl Cmd {
    fn new() -> Cmd {
        Cmd {
            exe: "".into(),
            params: vec![],
            current_dir: None,
            env: HashMap::new(),
        }
    }

    fn execute(&self) -> Result<()> {
        let mut cmd = Command::new(&self.exe);
        cmd.args(&self.params)
            .env_clear()
            .env("PATH", std::env!("PATH"))
            .envs(&self.env)
            // .stdout(cfg)
            // .stderr(cfg)
            ;

        if let Some(cdir) = &self.current_dir {
            cmd.current_dir(std::fs::canonicalize(cdir).unwrap());
        }

        let confirm = "[Y/n]? ";

        print!("Will execute {}{}", self, confirm);
        stdout().flush().unwrap();

        let mut yes_no = String::new();
        std::io::stdin()
            .read_line(&mut yes_no)
            .expect("Unable to read user’s input");

        if !yes_no.to_lowercase().starts_with('n') {
            if cmd.status().unwrap().success() {
                eprintln!("Youpi !");
                // Other checks?
            } else {
                eprintln!("Error executing the command")
            }
        }

        Ok(())
    }
}

impl fmt::Display for Cmd {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let command = if !self.params.is_empty() {
            format!("{} {}", &self.exe, &self.params.join(" "))
        } else {
            self.exe.clone()
        };
        write!(f, "`{}`", command)?;

        if let Some(cdir) = &self.current_dir {
            write!(f, " in {:?}", cdir)?;
        }

        if !self.env.is_empty() {
            writeln!(f, " with the following environment variable:")?;
            writeln!(f, "{:?}", self.env)
        } else {
            write!(f, " without any environment variable. ")
        }
    }
}
